﻿#region License Information

/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */

#endregion

using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Optimization;

namespace HeuristicLab.Algorithms.SimulatedAnnealing
{
    [Item("TemperatureController", "Decides based on information from the heatingoperator, whether to use cooling or heating and calls the appropriate operator.")]
    [StorableClass]
    public class TemperatureController : SingleSuccessorOperator
    {
        #region Strings
        private const string ReheatingOperatorName = "ReheatingOperator";
        private const string AnnealingOperatorName = "AnnealingOperator";
        private const string MaximumIterationsName = "MaximumIterations";
        private const string InitialTemperatureName = "InitialTemperature";
        private const string LowerTemperatureName = "LowerTemperature";
        private const string IterationsName = "Iterations";
        private const string TemperatureStartIndexName = "TemperatureStartIndex";
        private const string CoolingName = "Cooling";
        private const string StartTemperatureName = "StartTemperature";
        private const string EndTemperatureName = "EndTemperature";
        private const string TemperatureName = "Temperature";
        private const string IsAcceptedName = "IsAccepted";
        private const string ConsecutiveRejectedSolutionsCountName = "ConsecutiveRejectedSolutions";
        private const string AverageAcceptanceRatioName = "AverageAcceptanceRatio";
        private const string AcceptanceMemoryName = "AcceptanceMemory";
        private const string LastQualityName = "LastQuality";
        private const string UphillMovesMemoryName = "UphillMovesMemory";
        private const string MaximizationName = "Maximization";
        private const string MoveQualityName = "MoveQuality";
        private const string InitialAcceptanceRateName = "InitialAcceptanceRate";
        private const string ResultsName = "Results";
        private const string CalculatedInitialTemperatureName = "CalculatedInitialTemperature";

        #endregion

        #region Parameter Properties
        public ILookupParameter<DoubleValue> TemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[TemperatureName]; }
        }
        public ILookupParameter<DoubleValue> InitialTemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[InitialTemperatureName]; }
        }
        public IValueLookupParameter<DoubleValue> LowerTemperatureParameter
        {
            get { return (IValueLookupParameter<DoubleValue>)Parameters[LowerTemperatureName]; }
        }
        public ILookupParameter<DoubleValue> StartTemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[StartTemperatureName]; }
        }
        public ILookupParameter<DoubleValue> EndTemperatureParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[EndTemperatureName]; }
        }
        public ILookupParameter<IntValue> TemperatureStartIndexParameter
        {
            get { return (ILookupParameter<IntValue>)Parameters[TemperatureStartIndexName]; }
        }
        public ILookupParameter<BoolValue> CoolingParameter
        {
            get { return (ILookupParameter<BoolValue>)Parameters[CoolingName]; }
        }
        public ILookupParameter<IntValue> IterationsParameter
        {
            get { return (ILookupParameter<IntValue>)Parameters[IterationsName]; }
        }
        public IValueLookupParameter<IntValue> MaximumIterationsParameter
        {
            get { return (IValueLookupParameter<IntValue>)Parameters[MaximumIterationsName]; }
        }
        public IValueLookupParameter<IOperator> AnnealingOperatorParameter
        {
            get { return (IValueLookupParameter<IOperator>)Parameters[AnnealingOperatorName]; }
        }
        public IValueLookupParameter<IReheatingOperator> ReheatingOperatorParameter
        {
            get { return (IValueLookupParameter<IReheatingOperator>)Parameters[ReheatingOperatorName]; }
        }
        public ILookupParameter<BoolValue> IsAcceptedParameter
        {
            get { return (ILookupParameter<BoolValue>)Parameters[IsAcceptedName]; }
        }
        public ILookupParameter<ItemList<BoolValue>> AcceptanceMemoryParameter
        {
            get { return (ILookupParameter<ItemList<BoolValue>>)Parameters[AcceptanceMemoryName]; }
        }
        public ILookupParameter<ItemList<DoubleValue>> UphillMovesMemoryParameter
        {
            get { return (ILookupParameter<ItemList<DoubleValue>>)Parameters[UphillMovesMemoryName]; }
        }
        public ILookupParameter<DoubleValue> MoveQualityParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[MoveQualityName]; }
        }
        public ILookupParameter<DoubleValue> LastQualityParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[LastQualityName]; }
        }
        public IValueLookupParameter<BoolValue> MaximizationParameter
        {
            get { return (IValueLookupParameter<BoolValue>)Parameters[MaximizationName]; }
        }
        public ILookupParameter<DoubleValue> InitialAcceptanceRateParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[InitialAcceptanceRateName]; }
        }
        public ILookupParameter<IntValue> ConsecutiveRejectedSolutionsCountParameter
        {
            get { return (ILookupParameter<IntValue>)Parameters[ConsecutiveRejectedSolutionsCountName]; }
        }
        public ILookupParameter<DoubleValue> AverageAcceptanceRatioParameter
        {
            get { return (ILookupParameter<DoubleValue>)Parameters[AverageAcceptanceRatioName]; }
        }
        #endregion

        [StorableConstructor]
        protected TemperatureController(bool deserializing) : base(deserializing) { }
        protected TemperatureController(TemperatureController original, Cloner cloner) : base(original, cloner) { }
        public TemperatureController()
          : base()
        {
            Parameters.Add(new LookupParameter<DoubleValue>(TemperatureName, "The current temperature."));
            Parameters.Add(new ValueLookupParameter<DoubleValue>(LowerTemperatureName, "The lower bound of the temperature."));
            Parameters.Add(new LookupParameter<IntValue>(IterationsName, "The number of iterations."));
            Parameters.Add(new ValueLookupParameter<IntValue>(MaximumIterationsName, "The maximum number of iterations which should be processed."));
            Parameters.Add(new ValueLookupParameter<BoolValue>(MaximizationName, "True if the problem is a maximization problem, otherwise false."));
            Parameters.Add(new ValueLookupParameter<IOperator>(AnnealingOperatorName, "The operator that cools the temperature."));
            Parameters.Add(new ValueLookupParameter<IReheatingOperator>(ReheatingOperatorName, "The operator that reheats the temperature if necessary."));
            Parameters.Add(new LookupParameter<IntValue>(TemperatureStartIndexName, "The index where the annealing or heating was last changed."));
            Parameters.Add(new LookupParameter<BoolValue>(CoolingName, "True when the temperature should be cooled, false otherwise."));
            Parameters.Add(new LookupParameter<DoubleValue>(StartTemperatureName, "The temperature from which cooling or reheating should occur."));
            Parameters.Add(new LookupParameter<DoubleValue>(EndTemperatureName, "The temperature to which should be cooled or heated."));
            Parameters.Add(new LookupParameter<BoolValue>(IsAcceptedName, "Whether the move was accepted or not."));
            Parameters.Add(new LookupParameter<ItemList<BoolValue>>(AcceptanceMemoryName, "Memorizes the last N acceptance decisions."));
            Parameters.Add(new LookupParameter<DoubleValue>(InitialTemperatureName, "The initial temperature."));

            Parameters.Add(new LookupParameter<ItemList<DoubleValue>>(UphillMovesMemoryName, "Memorizes the last 100 uphill moves."));
            Parameters.Add(new LookupParameter<DoubleValue>(MoveQualityName, "The value which represents the quality of a move."));
            Parameters.Add(new LookupParameter<DoubleValue>(LastQualityName, "The value which represents the quality of the last move."));
            Parameters.Add(new LookupParameter<DoubleValue>(InitialAcceptanceRateName, "The initial acceptance rate of average-sized hills used to calculate the initial temperature."));
            Parameters.Add(new LookupParameter<IntValue>(ConsecutiveRejectedSolutionsCountName, "Amount of consecutive rejected solutions."));
            Parameters.Add(new LookupParameter<DoubleValue>(AverageAcceptanceRatioName, "Average acceptance over full acceptance memory."));

        }

        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new TemperatureController(this, cloner);
        }

        public override IOperation Apply()
        {
            var uphillMoves = UphillMovesMemoryParameter.ActualValue;
            if (!InitialTemperatureFound())
            {
                return findInitialTemperature();
            }else
            {
                return new OperationCollection {
                    ExecutionContext.CreateOperation(ReheatingOperatorParameter.ActualValue),
                    base.Apply()
                };
            }
        }

        private IOperation findInitialTemperature()
        {
            var currentQuality = MoveQualityParameter.ActualValue.Value;
            var lastQuality = LastQualityParameter.ActualValue.Value;
            var isMaximation = MaximizationParameter.ActualValue.Value;
            var hillMemory = UphillMovesMemoryParameter.ActualValue;
            
            if(lastQuality != -1)
            {
                double hillSize = CalculateHillSize(lastQuality, currentQuality, isMaximation);
                
                if(hillSize > 0)
                {
                    hillMemory.Add(new DoubleValue(hillSize));
                    if(InitialTemperatureFound())
                    {
                        var results = (ResultCollection)ExecutionContext.Parent.Scope.Variables[ResultsName].Value;
                        var initialTemperature = CalculateInitialTemperatureBasedOnAverageHillSize(hillMemory);
                        InitialTemperatureParameter.ActualValue.Value = initialTemperature;
                        TemperatureParameter.ActualValue.Value = initialTemperature;
                        TemperatureStartIndexParameter.ActualValue.Value = Math.Max(0, IterationsParameter.ActualValue.Value - 1);
                        StartTemperatureParameter.ActualValue.Value = TemperatureParameter.ActualValue.Value;
                        results.Add(new Result(CalculatedInitialTemperatureName, new DoubleValue(initialTemperature)));
                    }
                }
            }

            LastQualityParameter.ActualValue.Value = currentQuality;
            return base.Apply();
        }

        private double CalculateInitialTemperatureBasedOnAverageHillSize(ItemList<DoubleValue> hillMemory)
        {
            var averageHillSize = hillMemory.Average(x => x.Value);
            return -(averageHillSize / Math.Log(InitialAcceptanceRateParameter.ActualValue.Value));

        }

        private double CalculateHillSize(double lastQuality, double currentQuality, bool isMaximation)
        {
            return isMaximation ? lastQuality - currentQuality : currentQuality - lastQuality;
        }

        private bool InitialTemperatureFound()
        {
            return UphillMovesMemoryParameter.ActualValue.Count == 100;
        }
    }
}